import os
import re
import uuid
from multiprocessing import BoundedSemaphore
from pathlib import Path
from typing import List, Optional

import warnings
from cryptography.utils import CryptographyDeprecationWarning

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=CryptographyDeprecationWarning)
    import paramiko
from azure.mgmt.compute.models import ResourceIdentityType
from azure.core.exceptions import HttpResponseError

from skyplane import exceptions, key_root
from skyplane.compute.azure.azure_auth import AzureAuthentication
from skyplane.compute.azure.azure_server import AzureServer
from skyplane.compute.cloud_providers import CloudProvider
from skyplane.utils import logger
from skyplane.utils.timer import Timer


class AzureCloudProvider(CloudProvider):
    def __init__(self, key_root=key_root / "azure"):
        super().__init__()
        self.auth = AzureAuthentication()

        key_root.mkdir(parents=True, exist_ok=True)
        self.private_key_path = key_root / "azure_key"
        self.public_key_path = key_root / "azure_key.pub"

        self.provisioning_semaphore = BoundedSemaphore(5)

    @property
    def name(self):
        return "azure"

    @staticmethod
    def region_list():
        return AzureAuthentication.get_region_config()

    @staticmethod
    def lookup_continent(region: str) -> str:
        lookup_dict = {
            "oceania": {"australiaeast", "australiacentral", "australiasoutheast", "australiacentral2"},
            "asia": {
                "eastasia",
                "japaneast",
                "japanwest",
                "koreacentral",
                "koreasouth",
                "southeastasia",
                "southindia",
                "centralindia",
                "westindia",
                "jioindiacentral",
                "jioindiawest",
            },
            "north-america": {
                "canadacentral",
                "canadaeast",
                "centralus",
                "eastus",
                "eastus2",
                "northcentralus",
                "westus",
                "westus2",
                "westus3",
                "southcentralus",
                "westcentralus",
            },
            "south-america": {"brazilsouth", "brazilsoutheast"},
            "europe": {
                "francecentral",
                "germanywestcentral",
                "northeurope",
                "norwayeast",
                "swedencentral",
                "switzerlandnorth",
                "switzerlandwest",
                "westeurope",
                "uksouth",
                "ukwest",
                "francesouth",
                "germanynorth",
                "norwaywest",
            },
            "africa": {"southafricanorth", "southafricawest"},
            "middle-east": {"uaenorth", "uaecentral"},
        }
        for continent, regions in lookup_dict.items():
            if region in regions:
                return continent
        return "unknown"

    @staticmethod
    def lookup_valid_instance(region: str, instance_name: str) -> Optional[str]:
        sku_mapping = AzureAuthentication.get_sku_mapping()
        if instance_name in sku_mapping[region]:
            return instance_name
        match = re.match(r"^(?P<base_name>.*)_v(?P<version>\d+)$", instance_name)
        if match:
            base_name = match.group("base_name")
            for version in range(int(match.group("version")), 0, -1):
                test_instance_name = f"{base_name}_v{version}" if version > 1 else base_name
                if test_instance_name in sku_mapping[region]:
                    logger.fs.warning(
                        f"[azure] Instance {instance_name} not found in region {region} "
                        "but was able to find a similar instance {test_instance_name}."
                    )
                    return test_instance_name
        logger.error(f"[azure] Instance {instance_name} not found in region {region} and could not infer a similar instance name.")
        raise exceptions.InsufficientVCPUException(
            f"Azure instance type {instance_name} is not available in region {region}. "
            "Please contact Azure support to request capacity and then run `skyplane init --reinit-azure`."
        )

    @staticmethod
    def get_transfer_cost(src_key, dst_key, premium_tier=True):
        """Assumes <10TB transfer tier."""
        src_provider, src_region = src_key.split(":")
        dst_provider, dst_region = dst_key.split(":")
        assert src_provider == "azure"
        if not premium_tier:
            # TODO: tracked in https://github.com/parasj/skylark/issues/59
            return NotImplementedError()

        src_continent = AzureCloudProvider.lookup_continent(src_region)
        dst_continent = AzureCloudProvider.lookup_continent(dst_region)

        if dst_provider != "azure":  # internet transfer
            # From North America, Europe to any destination
            if src_continent in {"north-america", "europe"}:
                return 0.0875
            # From Asia (China excluded), Australia, MEA to any destination
            elif src_continent in {"asia", "oceania", "middle-east", "africa"}:
                return 0.12
            # From South America to any destination
            elif src_continent == "south-america":
                return 0.181
            else:
                raise ValueError(f"Unknown transfer cost for {src_key} -> {dst_key}")
        else:  # local transfer
            # intracontinental transfer
            if src_continent == dst_continent:
                # Between regions within North America, Between regions within Europe
                if src_continent in {"north-america", "europe"} and dst_continent in {"north-america", "europe"}:
                    return 0.02
                # Between regions within Asia, Between regions within Oceania, Between regions within Middle East and Africa
                elif src_continent in {"asia", "oceania", "middle-east", "africa"} and dst_continent in {
                    "asia",
                    "oceania",
                    "middle-east",
                    "africa",
                }:
                    return 0.08
                # Between regions within South America
                elif src_continent == "south-america" and dst_continent == "south-america":
                    return 0.181
                else:
                    raise ValueError(f"Unknown transfer cost for {src_key} -> {dst_key}")
            # intercontinental transfer
            else:
                # From North America to other continents, From Europe to other continents
                if src_continent in {"north-america", "europe"}:
                    return 0.05
                # From Asia to other continents, From Oceania to other continents, From Africa to other continents
                elif src_continent in {"asia", "oceania", "middle-east", "africa"}:
                    return 0.08
                # From South America to other continents
                elif src_continent == "south-america":
                    return 0.16
                else:
                    raise ValueError(f"Unknown transfer cost for {src_key} -> {dst_key}")

    def get_instance_list(self, region: str) -> List[AzureServer]:
        compute_client = self.auth.get_compute_client()
        server_list = []
        for vm in compute_client.virtual_machines.list(AzureServer.resource_group_name):
            if vm.tags.get("skyplane", None) == "true" and AzureServer.is_valid_vm_name(vm.name) and vm.location == region:
                name = AzureServer.base_name_from_vm_name(vm.name)
                s = AzureServer(name)
                if s.is_valid():
                    server_list.append(s)
                else:
                    logger.warning(
                        f"Warning: malformed Azure instance {name} found and ignored. "
                        "You should go to the Microsoft Azure portal, investigate this manually, "
                        "and delete any orphaned resources that may be allocated."
                    )
        return server_list

    # Copied from gcp_cloud_provider.py --- consolidate later?
    def create_ssh_key(self):
        private_key_path = Path(self.private_key_path)
        if not private_key_path.exists():
            private_key_path.parent.mkdir(parents=True, exist_ok=True)
            key = paramiko.RSAKey.generate(4096)
            key.write_private_key_file(self.private_key_path, password="skyplane")
            with open(self.public_key_path, "w") as f:
                f.write(f"{key.get_name()} {key.get_base64()}\n")

    def set_up_resource_group(self, clean_up_orphans=True):
        resource_client = self.auth.get_resource_client()
        if resource_client.resource_groups.check_existence(AzureServer.resource_group_name):
            # Resource group already exists.
            # Take this moment to search for orphaned resources and clean them up...
            network_client = self.auth.get_network_client()
            if clean_up_orphans:
                instances_to_terminate = []
                for vnet in network_client.virtual_networks.list(AzureServer.resource_group_name):
                    if vnet.tags.get("skyplane", None) == "true" and AzureServer.is_valid_vnet_name(vnet.name):
                        name = AzureServer.base_name_from_vnet_name(vnet.name)
                        s = AzureServer(name, assume_exists=False)
                        if not s.is_valid():
                            instances_to_terminate.append(s)

                if len(instances_to_terminate) > 0:
                    logger.warning(
                        f"Note: there are {len(instances_to_terminate)} orphaned resources "
                        f"in the Azure resource group {AzureServer.resource_group_name}. "
                        f"Manually delete them from the Azure console: {[i.name for i in instances_to_terminate]}"
                    )
            return
        rg_result = resource_client.resource_groups.create_or_update(
            AzureServer.resource_group_name, {"location": AzureServer.resource_group_location, "tags": {"skyplane": "true"}}
        )
        assert rg_result.name == AzureServer.resource_group_name

    # This code, along with some code in azure_server.py, is based on
    # https://github.com/ucbrise/mage-scripts/blob/main/azure_cloud.py.
    def provision_instance(self, location: str, vm_size: str, name: Optional[str] = None, uname: str = "skyplane") -> AzureServer:
        assert ":" not in location, "invalid colon in Azure location"

        if name is None:
            name = f"skyplane-azure-{str(uuid.uuid4()).replace('-', '')}"

        with open(os.path.expanduser(self.public_key_path)) as f:
            pub_key = f.read()

        # Prepare for making Microsoft Azure API calls
        compute_client = self.auth.get_compute_client()
        network_client = self.auth.get_network_client()

        # Use the common resource group for this instance
        resource_group = AzureServer.resource_group_name

        # TODO: On future requests to create resources, check if a resource
        # with that name already exists, and fail the operation if so

        with Timer("Creating Azure network"):
            # Create a Virtual Network for this instance
            poller = network_client.virtual_networks.begin_create_or_update(
                resource_group,
                AzureServer.vnet_name(name),
                {"location": location, "tags": {"skyplane": "true"}, "address_space": {"address_prefixes": ["10.0.0.0/24"]}},
            )
            poller.result()

            # Create a Network Security Group for this instance
            # NOTE: This is insecure. We should fix this soon.
            poller = network_client.network_security_groups.begin_create_or_update(
                resource_group,
                AzureServer.nsg_name(name),
                {
                    "location": location,
                    "tags": {"skyplane": "true"},
                    "security_rules": [
                        {
                            "name": name + "-allow-all",
                            "protocol": "*",
                            "source_port_range": "*",
                            "source_address_prefix": "*",
                            "destination_port_range": "*",
                            "destination_address_prefix": "*",
                            "access": "Allow",
                            "priority": 300,
                            "direction": "Inbound",
                        }
                        # Azure appears to add default rules for outbound connections
                    ],
                },
            )
            nsg_result = poller.result()

            # Create a subnet for this instance with the above Network Security Group
            subnet_poller = network_client.subnets.begin_create_or_update(
                resource_group,
                AzureServer.vnet_name(name),
                AzureServer.subnet_name(name),
                {"address_prefix": "10.0.0.0/26", "network_security_group": {"id": nsg_result.id}},
            )

            # Create a public IPv4 address for this instance
            ip_poller = network_client.public_ip_addresses.begin_create_or_update(
                resource_group,
                AzureServer.ip_name(name),
                {
                    "location": location,
                    "tags": {"skyplane": "true"},
                    "sku": {"name": "Standard"},
                    "public_ip_allocation_method": "Static",
                    "public_ip_address_version": "IPV4",
                },
            )

            subnet_result = subnet_poller.result()
            public_ip_result = ip_poller.result()

        with Timer("Creating Azure NIC"):
            # Create a NIC for this instance, with accelerated networking enabled
            poller = network_client.network_interfaces.begin_create_or_update(
                resource_group,
                AzureServer.nic_name(name),
                {
                    "location": location,
                    "tags": {"skyplane": "true"},
                    "ip_configurations": [
                        {"name": name + "-ip", "subnet": {"id": subnet_result.id}, "public_ip_address": {"id": public_ip_result.id}}
                    ],
                    "enable_accelerated_networking": True,
                },
            )
            nic_result = poller.result()

        # Create the VM
        with Timer("Creating Azure VM"):
            with self.provisioning_semaphore:
                try:
                    poller = compute_client.virtual_machines.begin_create_or_update(
                        resource_group,
                        AzureServer.vm_name(name),
                        {
                            "location": location,
                            "tags": {"skyplane": "true"},
                            "hardware_profile": {"vm_size": self.lookup_valid_instance(location, vm_size)},
                            "storage_profile": {
                                # "image_reference": {
                                #     "publisher": "canonical",
                                #     "offer": "0001-com-ubuntu-server-focal",
                                #     "sku": "20_04-lts",
                                #     "version": "latest",
                                # },
                                "image_reference": {
                                    "publisher": "microsoft-aks",
                                    "offer": "aks",
                                    "sku": "aks-engine-ubuntu-1804-202112",
                                    "version": "latest",
                                },
                                "os_disk": {"create_option": "FromImage", "delete_option": "Delete"},
                            },
                            "os_profile": {
                                "computer_name": AzureServer.vm_name(name),
                                "admin_username": uname,
                                "linux_configuration": {
                                    "disable_password_authentication": True,
                                    "ssh": {"public_keys": [{"path": f"/home/{uname}/.ssh/authorized_keys", "key_data": pub_key}]},
                                },
                            },
                            "network_profile": {"network_interfaces": [{"id": nic_result.id}]},
                            # give VM managed identity w/ system assigned identity
                            "identity": {"type": ResourceIdentityType.system_assigned},
                        },
                    )
                    vm_result = poller.result()
                    logger.fs.debug(f"Created Azure VM {vm_result.name} w/ system MSI principal_id = {vm_result.identity.principal_id}")
                except HttpResponseError as e:
                    if "ResourceQuotaExceeded" in str(e):
                        raise exceptions.InsufficientVCPUException(f"Got ResourceQuotaExceeded error in Azure region {location}") from e
                    elif "QuotaExceeded" in str(e):
                        raise exceptions.InsufficientVCPUException(f"Got QuotaExceeded error in Azure region {location}") from e
                    else:
                        raise

        server = AzureServer(name)
        server.wait_for_ssh_ready()
        return server



import os
from enum import Enum, auto
import sys

import typer
from rich import print as rprint
from skyplane import cloud_config, config_path
from skyplane.utils import logger

from . import usage_constants


class UsageStatsEnabledness(Enum):
    ENABLED_EXPLICITLY = auto()
    DISABLED_EXPLICITLY = auto()
    ENABLED_BY_DEFAULT = auto()


def _map_type(value, boolean):
    if value is None:
        return False
    if boolean:
        if value.lower() in ["true", "yes", "1"]:
            return True
    else:
        if value.lower() in ["false", "no", "0"]:
            return True
    return False


def usage_stats_enabledness():
    # environment vairable has higher priority
    usage_stats_enabled_env_var = os.getenv(usage_constants.USAGE_STATS_ENABLED_ENV_VAR)
    if _map_type(usage_stats_enabled_env_var, False):
        return UsageStatsEnabledness.DISABLED_EXPLICITLY
    elif _map_type(usage_stats_enabled_env_var, True):
        return UsageStatsEnabledness.ENABLED_EXPLICITLY
    elif usage_stats_enabled_env_var is not None:
        raise ValueError(
            f"Valid value for {usage_constants.USAGE_STATS_ENABLED_ENV_VAR} "
            f"env var is (false/no/0) or (true/yes/1), but got {usage_stats_enabled_env_var}"
        )
    # then check in the config file
    usage_stats_enabled_config_var = None
    # TODO: Check the correct error
    try:
        usage_stats_enabled_config_var = cloud_config.get_flag("usage_stats")
    except FileNotFoundError:
        pass
    except Exception as e:
        logger.debug(f"Failed to load usage stats config {e}")

    if not usage_stats_enabled_config_var:
        return UsageStatsEnabledness.DISABLED_EXPLICITLY
    elif usage_stats_enabled_config_var:
        return UsageStatsEnabledness.ENABLED_EXPLICITLY
    elif usage_stats_enabled_config_var is not None:
        raise ValueError(
            f"Valid value for 'usage_stats' in {config_path}" f" is (false/no/0) or (true/yes/1), but got {usage_stats_enabled_config_var}"
        )

    return UsageStatsEnabledness.ENABLED_BY_DEFAULT


def usage_stats_enabled():
    return usage_stats_enabledness() is not UsageStatsEnabledness.DISABLED_EXPLICITLY


def set_usage_stats_via_config(value):
    current_status = usage_stats_enabledness()
    if current_status is UsageStatsEnabledness.DISABLED_EXPLICITLY:
        if (isinstance(value, bool) and not value) or (isinstance(value, str) and _map_type(value, False)):
            rprint("Usage stats collection is already disabled.")
            rprint(usage_constants.USAGE_STATS_REENABLE_MESSAGE)
            return
    elif current_status is UsageStatsEnabledness.ENABLED_EXPLICITLY:
        if (isinstance(value, bool) and value) or (isinstance(value, str) and _map_type(value, True)):
            rprint("Usage stats collection is already enabled.")
            rprint(usage_constants.USAGE_STATS_REENABLED_MESSAGE)
            return

    if (isinstance(value, bool) and not value) or (isinstance(value, str) and _map_type(value, False)):
        answer = usage_stats_disabled_reconfirmation()
        if not answer:
            # Do nothing if the user confirms not to disable
            return
        else:
            rprint(usage_constants.USAGE_STATS_REENABLE_MESSAGE)

    try:
        cloud_config.set_flag("usage_stats", value)
    except Exception as e:
        raise Exception("Failed to enable/disable by writing to" f"{config_path}") from e

    if cloud_config.get_flag("usage_stats"):
        rprint(usage_constants.USAGE_STATS_REENABLED_MESSAGE)


# Redacted: Should set environment variable directly from command line
# def set_usage_stats_via_env_var(value):
#     if value != "0" and value != "1":
#         raise ValueError("Valid value for usage_stats_env_var is 0 or 1, but got " + str(value))
#     os.environ[usage_constants.USAGE_STATS_ENABLED_ENV_VAR] = value


def show_usage_stats_prompt():
    usage_stats_var = usage_stats_enabledness()
    if usage_stats_var is UsageStatsEnabledness.DISABLED_EXPLICITLY:
        rprint(usage_constants.USAGE_STATS_DISABLED_MESSAGE)
    elif usage_stats_var in [UsageStatsEnabledness.ENABLED_BY_DEFAULT, UsageStatsEnabledness.ENABLED_EXPLICITLY]:
        rprint(usage_constants.USAGE_STATS_ENABLED_MESSAGE)
    else:
        raise Exception("Prompt message unknown.")


def usage_stats_disabled_reconfirmation():
    """
    Ask for confirmation when the user decides to disable metrics collection.
    default: the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default) or "no".
    timeout: the number of seconds to wait for user to respond.
    """
    prompt = "Would you still like to opt out of sharing anonymous usage metrics?"
    rprint(usage_constants.USAGE_STATS_DISABLED_RECONFIRMATION_MESSAGE + "\n")
    return typer.confirm(prompt, default=False)



from functools import lru_cache

from azure.core.exceptions import ResourceExistsError

from skyplane import exceptions
from skyplane.compute.azure.azure_auth import AzureAuthentication
from skyplane.utils import logger


class AzureStorageAccountInterface:
    """Class to manage state for an Azure storage account. Note that storage account names are globally unique."""

    def __init__(self, account_name: str):
        self.auth = AzureAuthentication()
        self.account_name = account_name

    @lru_cache(maxsize=1)
    def storage_account_obj(self):
        sm_client = self.auth.get_storage_management_client()
        storage_accounts = sm_client.storage_accounts.list()
        for storage_account in storage_accounts:
            if storage_account.name == self.account_name:
                return storage_account
        raise exceptions.MissingBucketException(f"Storage account {self.account_name} not found")

    @property
    def azure_region(self):
        return self.storage_account_obj().location

    @property
    def storage_management_client(self):
        return self.auth.get_storage_management_client()

    def query_storage_account(self, storage_account_name):
        for account in self.storage_management_client.storage_accounts.list():
            if account.name == storage_account_name:
                return account
        raise ValueError(
            f"Storage account {storage_account_name} not found (found {[account.name for account in self.storage_management_client.storage_accounts.list()]})"
        )

    def storage_account_exists(self):
        for account in self.storage_management_client.storage_accounts.list():
            if account.name == self.account_name:
                return True
        return False

    def create_storage_account(self, azure_region, resource_group, tier="Premium_LRS"):
        try:
            operation = self.storage_management_client.storage_accounts.begin_create(
                resource_group, self.account_name, {"sku": {"name": tier}, "kind": "BlockBlobStorage", "location": azure_region}
            )
            operation.result()
        except ResourceExistsError as e:
            logger.warning(f"Unable to create storage account as it already exists: {e}")
        self.storage_account_obj.cache_clear()



import tempfile
from pathlib import Path
from typing import Optional

import typer

from skyplane.cli.common import print_header
from skyplane import skyplane_root
from skyplane.cli.cli_impl.cp_replicate import confirm_transfer, launch_replication_job
from skyplane.obj_store.object_store_interface import ObjectStoreObject
from skyplane.replicate.replication_plan import ReplicationTopology, ReplicationJob
from skyplane.utils import logger


def replicate_random(
    src_region: str,
    dst_region: str,
    inter_region: Optional[str] = typer.Argument(None),
    num_gateways: int = typer.Option(1, "--num-gateways", "-n", help="Number of gateways"),
    num_outgoing_connections: int = typer.Option(
        32, "--num-outgoing-connections", "-c", help="Number of outgoing connections between each gateway"
    ),
    total_transfer_size_mb: int = typer.Option(2048, "--size-total-mb", "-s", help="Total transfer size in MB."),
    chunk_size_mb: int = typer.Option(8, "--chunk-size-mb", help="Chunk size in MB."),
    use_bbr: bool = typer.Option(True, help="If true, will use BBR congestion control"),
    reuse_gateways: bool = False,
    debug: bool = False,
):
    """Replicate objects from remote object store to another remote object store."""
    print_header()

    if inter_region:
        assert inter_region not in [src_region, dst_region] and src_region != dst_region
        topo = ReplicationTopology()
        for i in range(num_gateways):
            topo.add_instance_instance_edge(src_region, i, inter_region, i, num_outgoing_connections)
            topo.add_instance_instance_edge(inter_region, i, dst_region, i, num_outgoing_connections)
    elif src_region == dst_region:
        typer.secho(
            "Replicate random doesn't support replicating to the same region as it tests inter-gateway networks.", fg="red", err=True
        )
        raise typer.Exit(code=1)
    else:
        topo = ReplicationTopology()
        for i in range(num_gateways):
            topo.add_instance_instance_edge(src_region, i, dst_region, i, num_outgoing_connections)

    if total_transfer_size_mb % chunk_size_mb != 0:
        logger.warning(f"total_transfer_size_mb ({total_transfer_size_mb}) is not a multiple of chunk_size_mb ({chunk_size_mb})")
    n_chunks = int(total_transfer_size_mb / chunk_size_mb)

    transfer_list = []
    for i in range(n_chunks):
        src_obj = ObjectStoreObject(src_region.split(":")[0], "", str(i))
        dst_obj = ObjectStoreObject(dst_region.split(":")[0], "", str(i))
        transfer_list.append((src_obj, dst_obj))

    job = ReplicationJob(
        source_region=topo.source_region(),
        source_bucket=None,
        dest_region=topo.sink_region(),
        dest_bucket=None,
        transfer_pairs=transfer_list,
        random_chunk_size_mb=total_transfer_size_mb // n_chunks,
    )
    confirm_transfer(topo=topo, job=job, ask_to_confirm_transfer=False)
    stats = launch_replication_job(
        topo=topo, job=job, debug=debug, reuse_gateways=reuse_gateways, use_bbr=use_bbr, use_compression=False, use_e2ee=True
    )
    return 0 if stats["success"] else 1


def replicate_random_solve(
    src_region: str,
    dst_region: str,
    inter_region: Optional[str] = typer.Argument(None),
    num_gateways: int = typer.Option(1, "--num-gateways", "-n", help="Number of gateways"),
    num_outgoing_connections: int = typer.Option(
        32, "--num-outgoing-connections", "-c", help="Number of outgoing connections between each gateway"
    ),
    total_transfer_size_mb: int = typer.Option(2048, "--size-total-mb", "-s", help="Total transfer size in MB."),
    chunk_size_mb: int = typer.Option(8, "--chunk-size-mb", help="Chunk size in MB."),
    use_bbr: bool = typer.Option(True, help="If true, will use BBR congestion control"),
    reuse_gateways: bool = False,
    solve: bool = typer.Option(False, help="If true, will use solver to optimize transfer, else direct path is chosen"),
    solver_required_throughput_gbits: float = typer.Option(2, help="Solver option: Required throughput in gbps."),
    solver_throughput_grid: Path = typer.Option(
        skyplane_root / "profiles" / "throughput.csv", "--throughput-grid", help="Throughput grid file"
    ),
    solver_verbose: bool = False,
    debug: bool = False,
):
    """Replicate objects from remote object store to another remote object store."""
    print_header()

    if solve:
        from skyplane.cli.cli_solver import solve_throughput  # lazy import due to pip dependencies

        with tempfile.NamedTemporaryFile(mode="w") as f:
            solve_throughput(
                src_region,
                dst_region,
                solver_required_throughput_gbits,
                gbyte_to_transfer=total_transfer_size_mb / 1024.0,
                max_instances=num_gateways,
                throughput_grid=solver_throughput_grid,
                solver_verbose=solver_verbose,
                out=Path(f.name),
            )
            topo = ReplicationTopology.from_json(Path(f.name).read_text())
    elif inter_region:
        assert inter_region not in [src_region, dst_region] and src_region != dst_region
        topo = ReplicationTopology()
        for i in range(num_gateways):
            topo.add_instance_instance_edge(src_region, i, inter_region, i, num_outgoing_connections)
            topo.add_instance_instance_edge(inter_region, i, dst_region, i, num_outgoing_connections)
    else:
        assert src_region != dst_region
        topo = ReplicationTopology()
        for i in range(num_gateways):
            topo.add_instance_instance_edge(src_region, i, dst_region, i, num_outgoing_connections)

    if total_transfer_size_mb % chunk_size_mb != 0:
        logger.warning(f"total_transfer_size_mb ({total_transfer_size_mb}) is not a multiple of chunk_size_mb ({chunk_size_mb})")
    n_chunks = int(total_transfer_size_mb / chunk_size_mb)

    transfer_list = []
    for i in range(n_chunks):
        src_obj = ObjectStoreObject(src_region.split(":")[0], "", str(i))
        dst_obj = ObjectStoreObject(dst_region.split(":")[0], "", str(i))
        transfer_list.append((src_obj, dst_obj))

    job = ReplicationJob(
        source_region=topo.source_region(),
        source_bucket=None,
        dest_region=topo.sink_region(),
        dest_bucket=None,
        transfer_pairs=transfer_list,
        random_chunk_size_mb=total_transfer_size_mb // n_chunks,
    )
    confirm_transfer(topo=topo, job=job, ask_to_confirm_transfer=False)
    stats = launch_replication_job(topo=topo, job=job, debug=debug, reuse_gateways=reuse_gateways, use_bbr=use_bbr, use_compression=False)
    return 0 if stats["success"] else 1
